/*
  fract.cpp

  julia and mandelbrot fractals (2d and 3d)

  Copyright (C) 2004, Patrick Rammelt <rammelt@cs.tu-berlin.de>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* ------------------------------------------------------------------------
   fract.cpp
   Patrick Rammelt
   ------------------------------------------------------------------------ */
#ifndef _FRACT_CPP
#define _FRACT_CPP

#include <math.h>


/* ------------------------------------------------------------------------
   COMPLEX NUMBERS (2D-NUMBERS)
   ------------------------------------------------------------------------ */
class complexNumber
{
protected:

  double r;  // real part
  double i;  // imaginary part

public:

  // create 0 / equal to other / with real & imaginary part
  complexNumber (void) { set(0,0); }
  complexNumber (double v) { set(v); }
  complexNumber (const class complexNumber &c) { set(c); }
  complexNumber (double realPart, double imagPart) { set(realPart,imagPart); }
  virtual ~complexNumber (void) {}

  // set from double value
  virtual inline complexNumber& set (double v) { 
    r = v; 
    i = 0; 
    return (*this);
  }

  // copy
  virtual inline complexNumber& set (const class complexNumber &c) { 
    r = c.r; 
    i = c.i; 
    return (*this);
  }

  // set number (real & imaginary part)
  virtual inline complexNumber& set (double realPart, double imagPart) { 
    r = realPart; 
    i = imagPart; 
    return (*this);
  }

  // test realness
  inline bool isReal (void) const { return (i==0); }

  // get real / imaginary part
  inline double real (void) const { return (r); }
  inline double imag (void) const { return (i); }

  // euclidean dist to origin / other complex Number
  inline double origin (void) const { 
    return ( sqrt(r*r + i*i) ); 
  }
  inline double dist (const class complexNumber &c) const { 
    return ( sqrt( (r-c.r)*(r-c.r) + 
		   (i-c.i)*(i-c.i)) ); 
  }

  // make equal to
  inline class complexNumber operator =  (const class complexNumber &c) { 
    return (set(c));    
  }

  //  add to this 
  inline class complexNumber operator += (const class complexNumber &c) { 
    return (set (r+c.r, 
		 i+c.i));
  }

  // subtract from this
  inline class complexNumber operator -= (const class complexNumber &c) { 
    return (set (r-c.r, 
		 i-c.i));    
  }

  // multiplicate by
  inline class complexNumber operator *= (const class complexNumber &c) {
    return (set ((r*c.r) - (i*c.i), 
		 (r*c.i) + (i*c.r)));
  }

  // devide by
  inline class complexNumber operator /= (const class complexNumber &c) { 
    *this *= c.reciprocal();
    return (*this);
  }

  // add / subtract / multiplicate / devide
  inline class complexNumber operator + (const class complexNumber &c) const {
    class complexNumber s (*this); 
    s += c; 
    return (s); 
  }
  inline class complexNumber operator - (const class complexNumber &c) const {
    class complexNumber s (*this); 
    s -= c; 
    return (s); 
  }
  inline class complexNumber operator * (const class complexNumber &c) const {
    class complexNumber s (*this); 
    s *= c; 
    return (s); 
  }
  inline class complexNumber operator / (const class complexNumber &c) const {
    class complexNumber s (*this); 
    s /= c; 
    return (s); 
  }

  // compare two numbers
  bool operator == (const class complexNumber &c) const { 
    return ((real()==c.real()) && 
	    (imag()==c.imag())); 
  }

  // absolute value |*this|
  inline double absolute (void) const { return (origin()); }

  // argument == phase
  inline double argument (void) const { return (atan(i/r)); }
  inline double phase    (void) const { return (argument()); }

  // conjugate
  inline class complexNumber conjugate (void) const {
    class complexNumber s (r, -i); return (s); 
  }

  // reciprocal (1/*this)
  inline class complexNumber reciprocal (void) const {
    const double f = (r*r + i*i);
    class complexNumber s (r/f, -(i/f)); 
    return (s); 
  }

  // sinus / sinus hyperbolicus
  inline class complexNumber sinus (void) const {
    class complexNumber s (sin(r)*cosh(i), 
			   cos(r)*sinh(i)); 
    return (s); 
  }
  inline class complexNumber sinusHyperbolicus (void) const {
    class complexNumber s (sinh(r)*cos(i), 
			   cosh(r)*sin(i)); 
    return (s); 
  }

  // cosinus / cosinus hyperbolicus
  inline class complexNumber cosinus (void) const {
    class complexNumber s (  cos(r)*cosh(i), 
			   -(sin(r)*sinh(i))); 
    return (s); 
  }
  inline class complexNumber cosinusHyperbolicus (void) const {
    class complexNumber s (cosh(r)*cos(i), 
			   sinh(r)*sin(i)); 
    return (s); 
  }

  // tangent / tangent hyperbolicus
  inline class complexNumber tangent (void) const {
    const double f = 2, t = cos(f*r) + cosh(f*i);
    class complexNumber s (sin(f*r)/t, 
			   sinh(f*i)/t); 
    return (s); 
  }
  inline class complexNumber tangentHyperbolicus (void) const {
    const double f = 2, t = cosh(f*r) + cos(f*i);
    class complexNumber s (sinh(f*r)/t, 
			   sin(f*i)/t); 
    return (s); 
  }

  // cotangent / cotangent hyperbolicus
  inline class complexNumber cotangent (void) const {
    const double f = 2, t = cosh(f*i) - cos(f*r);
    class complexNumber s (sin(f*r)/t, 
			   -(sinh(f*i)/t)); 
    return (s); 
  }
  inline class complexNumber cotangentHyperbolicus (void) const {
    const double f = 2, t = cosh(f*r) - cos(f*i);
    class complexNumber s (sinh(f*r)/t, 
			   sin(f*i)/t); 
    return (s); 
  }
  
  // exponent (exp) / logarithm naturalis (ln)
  inline class complexNumber exponent (void) const {
    const double t = exp(r);
    class complexNumber s (cos(i)*t, 
			   sin(i)*t); 
    return (s); 
  }
  inline class complexNumber logarithmNaturalis (void) const {
    class complexNumber s (log(absolute()), 
			   argument()); 
    return (s); 
  }

  // power (*this^p) / squareroot
  inline class complexNumber power (const class complexNumber &p) const {
    class complexNumber s ((p * logarithmNaturalis()).exponent()); 
    return (s); 
  }
  inline class complexNumber squareRoot (void) const {
    double a = argument()/(double)2;
    double b = sqrt(absolute());
    class complexNumber s (cos(a) * b, 
			   sin(a) * b); 
    return (s); 
  }
};


/* ------------------------------------------------------------------------
   QUATERNION NUMBERS (4D-NUMBERS) 
   ------------------------------------------------------------------------ */
class quaternion 
{
protected:

  double x;
  double y;
  double z;
  double w;

public:

  // create 0 / copy / with 4 parts a,b,c,d / from two complex numbers
  quaternion (void) { set (0,0,0,0); }
  quaternion (double a) { set (a); }
  quaternion (const class quaternion &q) { set (q); }
  quaternion (double a, double b, double c, double d) { set (a,b,c,d); }
  quaternion (const class complexNumber &c1, 
	      const class complexNumber &c2) { set (c1,c2); }
  virtual ~quaternion (void) {}

  // copy 
  virtual inline quaternion& set (const class quaternion &q) { 
    x = q.x; 
    y = q.y; 
    z = q.z; 
    w = q.w;     
    return (*this);
  }

  // set to a real number
  virtual inline quaternion& set (double a) {
    x = a;
    y = 0;
    z = 0;
    w = 0; 
    return (*this);
  }

  // set all four parts
  virtual inline quaternion& set (double a, double b, double c, double d) { 
    x = a; 
    y = b; 
    z = c; 
    w = d; 
    return (*this);
  }

  // set from two complex numbers
  virtual inline quaternion&  set (const class complexNumber &c1, 
				   const class complexNumber &c2) {
    x = (c1.real()+c2.real()) / (double)2;
    y = (c1.imag()+c2.imag()) / (double)2;
    z = (c1.imag()-c2.imag()) / (double)2;
    w = (c2.real()-c1.real()) / (double)2;
    return (*this);
  }

  // test realness
  inline bool isReal (void) const { return (y==0 && z==0 && w==0); }

  // get 4 parts
  inline double first  (void) const { return (x); }
  inline double second (void) const { return (y); }
  inline double third  (void) const { return (z); }
  inline double fourth (void) const { return (w); }

  // split into two complex numbers
  inline class complexNumber split1 (void) const {
    class complexNumber s (x-w,y+z); 
    return (s); 
  }
  inline class complexNumber split2 (void) const {
    class complexNumber s (x+w,y-z); 
    return (s); 
  }
  inline void split (class complexNumber &c1, 
		     class complexNumber &c2) const {
    c1.set (split1()); 
    c2.set (split2()); 
  }

  // euclidean distances to origin / other quaternion
  inline double origin (void) const { return (sqrt(x*x + y*y + z*z + w*w)); }
  inline double dist (const class quaternion &q) const { 
    return ( sqrt((x-q.x)*(x-q.x) + 
		  (y-q.y)*(y-q.y) +
		  (z-q.z)*(z-q.z) +
		  (w-q.w)*(w-q.w)) ); 
  }

  // make equal to
  inline quaternion& operator = (const class quaternion &q) { 
    return (set(q));
  }

  // add
  inline quaternion& operator += (const class quaternion &q) { 
    return (set (x+q.x, y+q.y, z+q.z, w+q.w));
  }

  // subtract
  inline quaternion& operator -= (const class quaternion &q) { 
    return (set (x-q.x, y-q.y, z-q.z, w-q.w));
  }

  // multipicate
  inline quaternion& operator *= (const class quaternion &q) { 
    return (set (x*q.x - y*q.y - z*q.z - w*q.w,
		 x*q.y + y*q.x + z*q.w - w*q.z,
		 x*q.z + z*q.x + w*q.y - y*q.w,
		 x*q.w + w*q.x + y*q.z - z*q.y));
  }

  // devide
  inline quaternion& operator /= (const class quaternion &q) {
    double xw = sqrt(q.x) - sqrt(q.w);
    double yz = sqrt(q.y) - sqrt(q.z);
    return (set ((((x*q.x)-(w*q.w)) / xw) + (((y*q.y)-(z*q.z)) / yz),
		 (((y*q.x)+(z*q.w)) / xw) - (((x*q.y)+(w*q.z)) / yz),
		 (((x*q.z)+(w*q.y)) / yz) + (((y*q.w)+(z*q.x)) / xw),
		 (((y*q.z)-(z*q.y)) / yz) - (((x*q.w)-(w*q.x)) / yz)));
  }

  // add / subtract / multiplicate / devide 
  inline class quaternion operator + (const class quaternion &c) const {
    class quaternion s (*this); 
    s += c; 
    return (s); 
  }
  inline class quaternion operator - (const class quaternion &c) const {
    class quaternion s (*this); 
    s -= c;
    return (s); 
  }
  inline class quaternion operator * (const class quaternion &c) const {
    class quaternion s (*this);
    s *= c;
    return (s); 
  }
  inline class quaternion operator / (const class quaternion &c) const {
    class quaternion s (*this);
    s /= c;
    return (s); 
  }

  // compare two numbers
  inline bool operator == (const class quaternion &c) const { 
    return ((first()==c.first()) && (second()==c.second()) && 
	    (third()==c.third()) && (fourth()==c.fourth())); 
  }

  // absolute value |*this|
  inline double absolute (void) const { 
    return (origin()); 
  }

  // conjugate
  inline class quaternion conjugate (void) const {
    class quaternion s (x, -y, -z, -w); 
    return (s); 
  }

  // sinus / sinus hyperbolicus
  inline class quaternion sinus(void) const {
    if (isReal()) {
      class quaternion q(sin(x));
      return (q);
    } else {
      double l = sqrt(y * y + z * z + w * w);
      double s = sin(x);
      double c = cos(x);
      double il = 1 / l;
      double e0 = exp(-l);
      double e1 = exp(l);
      
      double c0 = -0.5 * e0 * il * c;
      double c1 =  0.5 * e1 * il * c;
      
      class quaternion p(0.5 * e0 * s, c0 * y, c0 * z, c0 * w);
      class quaternion q(0.5 * e1 * s, c1 * y, c1 * z, c1 * w);
      return (p+q);
    }
  }

  // cosinus / cosinus hyperbolicus
  inline class quaternion cosinus (void) const {
    if (isReal()) {
      class quaternion q(sin(x));
      return (q);
    } else {
      double l = sqrt(y * y + z * z + w * w);
      double s = sin(x);
      double c = cos(x);
      double il = 1 / l;
      double e0 = exp(-l);
      double e1 = exp(l);
 
      double c0 =  0.5 * e0 * il * s;
      double c1 = -0.5 * e1 * il * s;
 
      class quaternion p(0.5 * e0 * c, c0 * y, c0 * z, c0 * w);
      class quaternion q(0.5 * e1 * c, c1 * y, c1 * z, c1 * w);
      return (p+q);
    }
  }

  // exponent (exp)
  inline class quaternion exponent (void) const {
    if (isReal()) {
      class quaternion q(exp(x));
      return (q);
    } else {
      double l = sqrt(y * y + z * z + w * w);
      double s = sin(l);
      double c = cos(l);
      double e = exp(x);
      double t = e * s / l;
      class quaternion q(e * c, t * y, t * z, t * w);
      return (q);
    }
  }

  // logarithm naturalis (ln)
  inline class quaternion logarithmNaturalis(void) const {
    if (isReal()) {
      if (x > 0) {
	class quaternion q(x);
	return (q);
      } else if (x < 0) {
	// The log of a negative purely real quaternion has
	// infinitely many values, all of the form (ln(-w), PI * I),
	// where I is any unit vector.  We arbitrarily choose an I
	// of (1, 0, 0) here and whereever else a similar choice is
	// necessary.  Geometrically, the set of roots is a sphere
	// of radius PI centered at ln(-w) on the real axis.
	class quaternion q(log(-x),M_PI,0,0);
	return (q);
      } else {
	// error . . . ln(0) not defined
	class quaternion q(0);
	return (q);
      }
    } else {
      double l = sqrt(y * y + z * z + w * w);
      double r = sqrt(l * l + x * x);
      double theta = atan2(l, x);
      double t = theta / l;
      class quaternion q(log(r), t * y, t * z, t * w);
      return (q);
    }
  }

  // power (*this^p) 
  inline class quaternion power (const class quaternion &p) const {
    class quaternion s ((p * logarithmNaturalis()).exponent()); return (s); 
  }

  // square root
  inline quaternion squareRoot (void) const {
    // In general, the square root of a quaternion has two values, one
    // of which is the negative of the other.  However, any negative purely
    // real quaternion has an infinite number of square roots.
    // This function returns the positive root for positive reals and
    // the root on the positive i axis for negative reals.
    if (isReal()) {
      if (x >= 0) {
	class quaternion q(sqrt(x), 0, 0, 0);
	return (q);
      } else {
	class quaternion q(0, sqrt(-x), 0, 0);
	return (q);
      }
    } else {
      double b = sqrt(y * y + z * z + w * w);
      double r = sqrt(x * x + b * b);
      if (x >= 0) {
	double m = sqrt(0.5 * (r + x));
	double l = b / (2 * m);
	double t = l / b;
	class quaternion q(m, y * t, z * t, w * t);
	return (q);
      } else {
	double l = sqrt(0.5 * (r - x));
	double m = b / (2 * l);
	double t = l / b;
	class quaternion q(m, y * t, z * t, w * t);
	return (q);
      }
    }
  }

  // sinus / sinus hyperbolicus on split-objects
  inline class quaternion splitSin (void) const {
    class quaternion s(split1().sinus(), split2().sinus()); 
    return (s);
  }
  inline class quaternion splitSinH (void) const {
    class quaternion s(split1().sinusHyperbolicus(),
		       split2().sinusHyperbolicus()); 
    return (s);
  }

  // cosinus / cosinus hyperbolicus on split-objects
  inline class quaternion splitCos (void) const {
    class quaternion s(split1().cosinus(), split2().cosinus()); 
    return (s);
  }
  inline class quaternion splitCosH (void) const {
    class quaternion s(split1().cosinusHyperbolicus(),
		       split2().cosinusHyperbolicus()); 
    return (s);
  }

  // cotangent / cotangent hyperbolicus on split-objects
  inline class quaternion splitCTan (void) const {
    class quaternion s(split1().cotangent(), 
		       split2().cotangent()); 
    return (s);
  }
  inline class quaternion splitCTanH (void) const {
    class quaternion s(split1().cotangentHyperbolicus(),
		       split2().cotangentHyperbolicus()); 
    return (s);
  }

  // tangent / tangent hyperbolicus on split-objects
  inline class quaternion splitTan (void) const {
    class quaternion s(split1().tangent(), split2().tangent()); 
    return (s);
  }
  inline class quaternion splitTanH (void) const {
    class quaternion s(split1().tangentHyperbolicus(),
		       split2().tangentHyperbolicus()); 
    return (s);
  }

  // exp, logarithm naturalis, square root on slpit-objects
  inline class quaternion splitExp (void) const {
    class quaternion s(split1().exponent(), split2().exponent()); 
    return (s);
  }
  inline class quaternion splitLN (void) const {
    class quaternion s(split1().logarithmNaturalis(), 
		       split2().logarithmNaturalis()); 
    return (s);
  }
  inline class quaternion splitSqrt (void) const {
    class quaternion s(split1().squareRoot(), 
  		       split2().squareRoot()); 
    return (s);
  }

};


/* ------------------------------------------------------------------------
   MANDELBROT FRACTAL 
   ------------------------------------------------------------------------ */
template <class numbertype>
class mandelbrot
{
protected:

  double bout;   // bailout - stop if dist of z to origin greater bout
  long   maxit;  // max iteration - stop if iterations >= maxit

public:

  // create default / with special bailout and max. iterations
  mandelbrot (void) { 
    set();
  }
  mandelbrot (double bailout, long maxiter) { 
    set (bailout, maxiter);
  }
  virtual ~mandelbrot (void) { }

  // set default values
  virtual inline mandelbrot& set (void) { 
    return (set (2.0, 50)); 
  }

  // set bailout / max. iterations 
  virtual inline mandelbrot& set (double bailout, long maxiter) { 
    this->bout  = bailout; 
    this->maxit = maxiter; 
    return (*this);
  }

  // get bailout / max. iterations 
  inline double bailout (void) const { return (this->bout);  }
  inline long   maxiter (void) const { return (this->maxit); }

  // eval point c (return number of steps)
  virtual inline long eval (const numbertype &c) const { 
    numbertype z; 
    return (eval (c,z,0)); 
  }
 
  // distance to origin when iteration stops
  virtual inline numbertype endPoint (const numbertype &c) const { 
    numbertype z; 
    return (endPoint (c,z,0)); 
  }

  // cumulated length of the "way" from start to endpoint
  virtual inline numbertype pathLength (const numbertype &c) const { 
    numbertype z; 
    return (pathLength (c,z,0)); 
  }

  // is point c part of mandelbrot set (eval >= max.iterations)
  virtual inline bool partof (const numbertype &c) const { 
    return ((eval(c)>=this->maxit)); 
  }
    
protected:

  // iterating z
  virtual inline void iterate (const numbertype &c, numbertype &z) const { 
    z = (z*z) + c;
  }

  // number of steps
  virtual inline long eval (const numbertype &c, 
			    const numbertype &z, long n) const {
    for (numbertype zi (z); n<this->maxit; iterate (c,zi), n++) 
      if (zi.origin()>=this->bout) return (n);
    return (this->maxit); 
  }

  // recursive version
  // virtual inline long eval (const numbertype &c, 
  //                                 numbertype &z, long n) const {
  //   if (n>=this->maxit || z.origin()>=this->bout) return (n);
  //   else return ( eval (c, (z*z)+c, n+1) ); 
  // } 

  // point where iteration stops
  virtual inline numbertype endPoint (const numbertype &c, 
				      const numbertype &z, long n) const {
    numbertype zi(z);
    for ( ; n<this->maxit; iterate (c,zi), n++) 
      if (zi.origin()>=this->bout) return (zi);
    return (zi); 
  }

  // cumulated length of the "way" from start to endpoint
  virtual inline numbertype pathLength (const numbertype &c, 
					const numbertype &z, long n) const {
    numbertype zj, zi(z), d;
    for ( ; n<this->maxit; n++) {
      if (zi.origin()>=this->bout) return (d);
      zj = zi; iterate (c,zi); d += (zi-zj);
    }
    return (d); 
  }
};


/* ------------------------------------------------------------------------
   JULIA FRACTAL 
   ------------------------------------------------------------------------ */
template <class numbertype>
class julia : public mandelbrot<numbertype>
{
public:

  // create default / with special bailout and max. iterations
  julia (void) { 
    this->set ();
  }
  julia (double bailout, long maxiter) { 
    this->set (bailout, maxiter); 
  }
  virtual ~julia (void) { }

  // eval point z of mandelbrot point c (return number of steps)
  virtual inline long eval (const numbertype &c, const numbertype &z) const {
    return (mandelbrot<numbertype>::eval(c,z,0)); 
  }

  // eval point z of mandelbrot point c (return endpoint point)
  virtual inline numbertype endPoint (const numbertype &c, 
				      const numbertype &z) const {
    return (mandelbrot<numbertype>::endPoint(c,z,0)); 
  }

  // cumulated length of the "way" from start to endpoint
  virtual inline numbertype pathLength (const numbertype &c, 
					 const numbertype &z) const {
    return (mandelbrot<numbertype>::pathLength(c,z,0)); 
  }

  // is point z part of mandelbrot set (eval >= max.iterations)
  virtual inline bool partof (const numbertype &c, const numbertype &z) const {
    return ((eval(c,z)>=this->maxit)); 
  }
};


#endif // ifndef _FRACT_CPP
